/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
/******************************************************************************
 *  drawtool.c   drawing routines                                            *
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>

#include <assert.h>

#include "vector.h"
#include "units.h"

long sumsize, objsize;

#define MAXNEEDED (1024*256)
    
extern XFontStruct *vsmallFont;

GC brushedGC, tiledGC;
extern GC uniGC;
static char *buffer;
long lastmax;
int lastj;


/***** Draw entire map *****/
void DrawMap(MapStruct * ms, Pixmap image)
{
    int i, grid = 0;

    buffer = (char *) malloc(MAXNEEDED);
    /* Setup Projection */
    map_params(ms);

    /* Read Configuration File */
    if (ms->theme.on)
	ReadConfig(ms, "thematic.con");
    else
	ReadConfig(ms, "mapnew.con");

    /* Prepare Output Image */
    if (image) {
	/* Assign background/foreground colors (white/black) */
	ImageColorAllocate(image, 0xffffff);
	ImageColorAllocate(image, 0);

	/* Allocate Feature Colors and Brushes */
	for (i = 0; i < ms->ndirs; i++) {
	    if (ms->datasets[i].on > 0) {
		if (!strcmp(ms->datasets[i].name, "GRID"))
		    grid = 1;
		SetColor(image, &ms->datasets[i].line);
		SetColor(image, &ms->datasets[i].fill);
		SetColor(image, &ms->datasets[i].point);
	    }
	}
	for (i = 0; i < ms->nlyrs; i++) {
	    if (ms->layers[i].on > 0) {
		SetColor(image, &ms->layers[i].line);
		SetColor(image, &ms->layers[i].fill);
		SetColor(image, &ms->layers[i].point);
	    }
	}
	for (i = 0; i < ms->nfeats; i++) {
	    if (ms->features[i].on > 0) {
		SetColor(image, &ms->features[i].line);
		SetColor(image, &ms->features[i].fill);
		SetColor(image, &ms->features[i].point);
	    }
	}
    }
    /* cycle through other layers */
    DrawDatasets(image, ms);
    if (grid) LabelGraticule(image, ms, uniGC);
    free(buffer);
}


void chsize(long size, int line)
{
	static long lasts = 0;
#if 0
	if (size>lasts) {
		fprintf(stderr,"msize=%ld\n",size);
		lasts=size;
	}
#endif
	if (size>MAXNEEDED && size+100>lasts) {
//		fprintf(stderr,"Had to get more size in line %d (%ld)!\n",line, size);
		free(buffer);
		lasts = size+(size/20);
		buffer = malloc(lasts);
	}
}


/* scroll through available datasets, drawing the ones which may be visible */
void DrawDatasets(image, ms)
Pixmap image;
MapStruct *ms;
{
    int i;
    fpos_t offset = 32;
    char stidx[50];
    char ctbase[50], dir[60], mbrfn[70];
    char st[30], line[90];
    DIR *counties;
    FILE *states, *lyrfile;
    MBR bounds;
    char stfn[5], ctfn[5];
    struct stat fstat;
    struct dirent *county;
    char bufp[256];

    for (i = 0; i < ms->ndirs; i++) {
	if (ms->datasets[i].on > 0) {
	    if (!strcmp(ms->datasets[i].name, "GRID") && qconf_graticule()) {
		DrawGraticule(image, ms, ms->datasets[i].line.brush);
		continue;
	    } else if (!strcmp(ms->datasets[i].name, "BACK")) {
		if (ms->maxWidthD == 360)
		    DrawGlobe(image, ms, ms->datasets[i].fill.brush);
#if 0
		else
		    ImageFilledRectangle(image, 0, 0, ms->WidthP, ms->HeightP,
					 ms->datasets[i].fill.brush);
#endif
	    } else if (!strcmp(ms->datasets[i].name, "CITIES")) {
		LabelCities(image, ms, &ms->datasets[i]);
	    } else if (!strcmp(ms->datasets[i].name, "USA")) {
		sprintf(dir, "%s/USA/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "MEXICO")) {
		sprintf(dir, "%s/mexico/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "CENTRALAM")) {
		sprintf(dir, "%s/centralam/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "SOUTHAM")) {
		sprintf(dir, "%s/southam/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "CANADA")) {
		sprintf(dir, "%s/canada/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "SASAUS")) {
		sprintf(dir, "%s/sasaus/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "ANTARTICA")) {
		sprintf(dir, "%s/antartica/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "EURNASIA")) {
		sprintf(dir, "%s/eurnasia/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "AFRICA")) {
		sprintf(dir, "%s/africa/",get_vmap_root());
		DrawLayers(image, ms, dir);
	    } else if (!strcmp(ms->datasets[i].name, "*")) {
		sprintf(stidx,"%s/%s",get_vmap_root(),"index");
		if ((states = fopen(stidx, "rt")) == NULL) {
		    sprintf(bufp,"DrawDatasets: %s",stidx);
#if 1 // makes it annoying to turn off gps tracking if dialog keeps appearing
		    fprintf(stderr, "%s\n", bufp); /* XXX */
#else
		    error(bufp,errno);
#endif
		    continue;
		}
#if 1
		while (fscanf(states, "%[^\n]\n", line) != EOF) {
		    if (sscanf(line, "%ld %ld %ld %ld %2[0-9] %[A-Z ]", &bounds.WLon,
			       &bounds.SLat, &bounds.ELon, &bounds.NLat, stfn, st) < 6) {
			fprintf(stderr, "  invalid index line: %s\n", line);
			continue;
		    }
		    if (!OverlapMBR(bounds, ms->BoundsU))
			continue;
		    sprintf(ctbase, "%s/%s/", get_vmap_root(), stfn);
		    if (stat(ctbase, &fstat))
			continue;
		    if ((counties = opendir(ctbase)) == NULL) {
			sprintf(bufp,"DrawDatasets: %s",ctbase);
			error(bufp,errno);
		    }
		    while ((county = readdir(counties)) != NULL) {
			if (!sscanf(county->d_name, "%[0-9]", ctfn))
			    continue;
			sprintf(dir, "%s%s/", ctbase, ctfn);
			sprintf(mbrfn, "%splaces.lyr", dir);
			if ((lyrfile = fopen(mbrfn, "rb")) == NULL) {
			    sprintf(bufp,"DrawDatasets: %s",mbrfn);
			    error(bufp,errno);
			}
			fsetpos(lyrfile, &offset);
			fread(&bounds, 4, 4, lyrfile);
			bounds.WLon = htonl(bounds.WLon);
			bounds.SLat = htonl(bounds.SLat);
			bounds.ELon = htonl(bounds.ELon);
			bounds.NLat = htonl(bounds.NLat);
			fclose(lyrfile);
			CenterMBR(&bounds);
			if (!OverlapMBR(bounds, ms->BoundsU))
			    continue;
			DrawLayers(image, ms, dir);
		    }
		    closedir(counties);
		}
#endif
		fclose(states);
	    }
	}
    }
}

/* search through all layers (files) in a particular dataset */
void DrawLayers(image, ms, dir)
Pixmap image;
MapStruct *ms;
char *dir;
{
    char lyrpath[60], basepath[60];
    struct stat fstat;
    short i = 0;

    for (i = 0; i < MAXLAYERS; i++) {
	if (ms->layers[i].on > 0) {
	    strcpy(basepath, dir);
	    strcat(basepath, ms->layers[i].name);
	    strcpy(lyrpath, basepath);
	    strcat(lyrpath, ".lyr");
	    if (stat(lyrpath, &fstat))
		continue;
	    DrawObjects(image, ms, basepath, &ms->layers[i]);
	}
    }
}

/* search through all objects in a particular layer */
void DrawObjects(image, ms, basename, config)
Pixmap image;
MapStruct *ms;
char *basename;
Config *config;
{
    Symbol *linesym, *fillsym, *ptsym, *oldline = NULL, *oldfill = NULL, *oldpt = NULL;
    long i = 0;
    char *extsum = ".sum", *extobj = ".obj", oldcfcc[4] = " ";
    char sum[70], obj[70];
    SummaryRec *s, *p, q;
    char *sumfile, *objfile;
    int fdsum, fdobj, fnum;
    struct stat statbuf;
    char bufp[256];

    strcpy(sum, basename);
    strcpy(obj, basename);
    strcat(sum, extsum);

    if ((fdsum = open(sum, O_RDONLY)) == -1) {
	sprintf(bufp,"DrawObjects: %s",sum);
	error(bufp,errno);
    }
    fstat(fdsum,&statbuf);
    sumsize = statbuf.st_size;
    sumfile = mmap(0, sumsize, PROT_READ, MAP_SHARED, fdsum, 0);

    strcat(obj, extobj);

    if ((fdobj = open(obj, O_RDONLY)) == -1) {
	sprintf(bufp,"DrawObjects: %s",obj);
	error(bufp,errno);
    }
    fstat(fdobj,&statbuf);
    objsize = statbuf.st_size;
    objfile = mmap(0, objsize, PROT_READ, MAP_SHARED, fdobj, 0);

    if (config->line.color != -1)
	linesym = &config->line;
    else
	linesym = NULL;

    if (config->fill.color != -1)
	fillsym = &config->fill;
    else
	fillsym = NULL;

    if (config->point.color != -1)
	ptsym = &config->point;
    else
	ptsym = NULL;

    for (p=(SummaryRec *)sumfile; (void *)p<(void *)(sumfile+sumsize); p++) {
	i++;

	s = &q;
	memcpy(s, p, sizeof(q));
	s->offset = htonl(s->offset);
	s->bounds.WLon = htonl(s->bounds.WLon);
	s->bounds.SLat = htonl(s->bounds.SLat);
	s->bounds.ELon = htonl(s->bounds.ELon);
	s->bounds.NLat = htonl(s->bounds.NLat);

	if (s->offset < 0) {	/* Deleted record or alias? -- ignore */
	    continue;
	}
	CenterMBR(&s->bounds);
	if (OverlapMBR(s->bounds, ms->BoundsU)) {
	    FGetObj(objfile, &ms->object, s->offset);
	    if (strcmp(ms->object.cfcc, oldcfcc)) {
		fnum = FeatureSym(ms, &linesym, &fillsym, &ptsym, ms->object.cfcc);
		if (fnum) {
		    if (ms->features[fnum - 1].on == 0)
			continue;
		} else {
		    if (config->line.color != -1)
			linesym = &config->line;
		    else
			linesym = NULL;
		    if (config->fill.color != -1)
			fillsym = &config->fill;
		    else
			fillsym = NULL;
		    if (config->point.color != -1)
			ptsym = &config->point;
		    else
			ptsym = NULL;
		}
	    } else {
		linesym = oldline;
		fillsym = oldfill;
		ptsym = oldpt;
	    }
	    switch (ms->object.type) {
	    case LINESTRING:
		if (linesym != NULL)
		    DrawPolyline(image, ms, objfile, s, linesym);
		break;
	    case POLYGON:
		if ((linesym != NULL) || (fillsym != NULL))
		    DrawPolygon(image, ms, objfile, s, linesym, fillsym);
		break;
	    case POINT:
		if (ptsym != NULL)
		    DrawPoint(image, ms, s, ptsym);
		break;
	    default:
		fprintf(stderr, "%s(%i):", config->name, ms->object.type);
	    }
	    strcpy(oldcfcc, ms->object.cfcc);
	    oldline = linesym;
	    oldfill = fillsym;
	    oldpt = ptsym;
	}
    }
    munmap(sumfile, sumsize);
    munmap(objfile, objsize);
    close(fdsum);
    close(fdobj);
}

/* Draw a single linestring feature */
void DrawPolyline(image, ms, objfile, s, style)
Pixmap image;
MapStruct *ms;
char *objfile;
SummaryRec *s;
Symbol *style;
{
    long size, off2, npts, offset2, lat, lon, lastlat, lastlon;
    int x, y, lastx, lasty, i;
    char action, allinside, sector, lastsect, lastplot;

    offset2 = ms->object.offset2;
    FGetObjNPts(objfile, offset2, &npts);

    npts = htonl(npts);

    off2 = offset2 + 6;
    size = ms->object.length + 4 - (off2 - s->offset);

    chsize(size,__LINE__);

    FGetObjPtBuf(objfile, off2, size, buffer);
    offset2 = 0;
    allinside = RectInRect(s->bounds, ms->BoundsU);

    for (i = 0; i < npts; i++) {
	FGetObjPtfromBuf(buffer, &offset2, &action, &lat, &lon);
	lon = SubMod(lon - DEG180_LU, ms->CtrLonU);
	lat -= DEG90_LU;
	if (!allinside)
	    sector = Sector(lon, lat, ms->nBoundsU);
	if ((action == TO) && (allinside || !(sector & lastsect))) {
	    if (!lastplot) {
		nProject(ms, lastlat, lastlon, &lastx, &lasty);
	    }
	    nProject(ms, lat, lon, &x, &y);
	    if ((x != lastx) || (y != lasty)) {
		ImageLine(image, lastx, lasty, x, y, style->brush);
		lastx = x;
		lasty = y;
	    }
	    lastplot = TRUE;
	} else
	    lastplot = FALSE;
	lastsect = sector;
	lastlat = lat;
	lastlon = lon;
    }
}

/* Draw a single (possibly filled) polygon feature */
void DrawPolygon(image, ms, objfile, sum, linesym, fillsym)
Pixmap image;
MapStruct *ms;
char *objfile;
SummaryRec *sum;
Symbol *linesym, *fillsym;
{
    long size, off2, npts, offset2, lat, lon, unlon, lastlat,
     lastlon;
    int x, y, lastx, lasty, i, j, wd = ms->WidthP, ht = ms->HeightP;
    char action, allinside, firstin, sector, lastsect, lastplot;

    offset2 = ms->object.offset2;
    FGetObjNPts(objfile, offset2, &npts);

    npts = htonl(npts);

    off2 = offset2 + 6;
    size = ms->object.length + 4 - (off2 - sum->offset);

    chsize(size,__LINE__);

    FGetObjPtBuf(objfile, off2, size, buffer);
    offset2 = 0;
    j = 0;
    allinside = RectInRect(sum->bounds, ms->BoundsU);
    lastx = 0;
    lasty = 0;
    firstin = FALSE;
    lastplot = FALSE;
    for (i = 0; i < npts; i++) {
	FGetObjPtfromBuf(buffer, &offset2, &action, &lat, &unlon);
	lon = SubMod(unlon - DEG180_LU, ms->CtrLonU);
	lat -= DEG90_LU;
	if (!allinside)
	    sector = Sector(lon, lat, ms->nBoundsU);
	else {
	    sector = 0;
	    lastsect = 0;
	}
	if (action == FROM) {
	    if (firstin && !lastplot) {
		nProject(ms, lastlat, lastlon, &x, &y);
		ms->pointlist[j].x = x;
		ms->pointlist[j++].y = y;
		assert(j<=MAXPOINTS);
	    }
	    if (j > 2) {
		if (fillsym != NULL) {
			ImageFilledPolygon(image, ms->pointlist, j, fillsym->brush);
		}
		if (linesym != NULL) {
			ImagePolygon(image, ms->pointlist, j, linesym->brush);
		}
	    }
	    j = 0;
	    lastx = 0;
	    lasty = 0;
	}
	if (!allinside && (action == TO) && !(sector & lastsect) && !lastplot) {
	    nProject(ms, lastlat, lastlon, &lastx, &lasty);
	    ms->pointlist[j].x = lastx;
	    ms->pointlist[j++].y = lasty;
	    assert(j<=MAXPOINTS);
	}
	if (allinside || !sector || ((action == TO) && (sector != lastsect))) {
	    if (sector & lastsect) {
		x = (wd >> 1) + wd * ((sector >> 1 & 0x1) - (sector & 0x1));
		y = (ht >> 1) + ht * ((sector >> 3 & 0x1) - (sector >> 2 & 0x1));
	    } else {
		nProject(ms, lat, lon, &x, &y);
	    }
	    if ((x != lastx) || (y != lasty)) {
		ms->pointlist[j].x = x;
		ms->pointlist[j++].y = y;
		assert(j<=MAXPOINTS);
	    }
	    lastplot = TRUE;
	} else {
	    lastplot = FALSE;
	}
	lastsect = sector;
	lastlat = lat;
	lastlon = lon;
	lastx = ms->pointlist[j - 1].x;
	lasty = ms->pointlist[j - 1].y;
	if (action == FROM)
	    firstin = lastplot;
    }
    if (!lastplot && firstin) {
	Project(ms, lastlat, lastlon, &x, &y);
	ms->pointlist[j].x = x;
	ms->pointlist[j++].y = y;
	assert(j<=MAXPOINTS);
    }
    if (j > 2) {
	if (fillsym != NULL) 
		ImageFilledPolygon(image, ms->pointlist, j, fillsym->brush);
	if (linesym != NULL) 
		ImagePolygon(image, ms->pointlist, j, linesym->brush);
    }
}

void DrawPoint(Pixmap image, MapStruct *ms, SummaryRec *s, Symbol *style)
{
    int x, y;
    Project(ms, s->bounds.SLat, s->bounds.WLon, &x, &y);
	ImageLine(image, x, y, x, y, style->brush);
}

/***** draw the gridlines *****/

void DrawGraticule(image, ms, gc)
Pixmap image;
MapStruct *ms;
GC gc;
{
    long GridU, ThinU;		/* Grid Spacing and meridian vertex spacing */
    long xu, yu, start, end;
    int x1, x2, y, y2;
    float GridD;

    if ((ms->GridSpace < DEF_GRID) || (!ms->HalfWidU)) return;

    if (ms->GridSpace == DEF_GRID)
	ms->GridSpace = DEF_GRID_SP;

    GridD = ms->GridSpace / PIX_PER_DEG(ms);

    if (GridD >= 12)
	GridD = 15;
    else if (GridD >= 8)
	GridD = 10;
    else if (GridD >= 3)
	GridD = 5;
    else if (GridD >= 1.5)
	GridD = 2;
    else if (GridD >= .75)
	GridD = 1;
    else if (GridD >= .35)
	GridD = .5;
    else if (GridD >= .15)
	GridD = .2;
    else
	GridD = .1;

    GridU = DEG_TO_LU(GridD);
    ThinU = GridU / 3;

    /* Draw Parallels. */
    x1 = MU_TO_PIX(ms, ms->HalfWidU - ms->HalfWidU);
    x2 = MU_TO_PIX(ms, ms->HalfWidU + ms->HalfWidU) - 1;

    start = ms->BoundsU.SLat - (ms->BoundsU.SLat % GridU);

    for (yu = start; yu < ms->BoundsU.NLat; yu += GridU) {
	if (ms->maxWidthD == 360) {
	    Project(ms, yu, ms->BoundsU.WLon, &x1, &y);
	    x2 = ms->WidthP - x1;
	} else
	    y = MU_TO_PIX(ms, ms->CtrLatU - yu) + ms->HeightP / 2;
	ImageLine(image, x1, y, x2, y, gc);
    }

    /* Draw Meridians */
    start = ms->BoundsU.WLon - (ms->BoundsU.WLon % GridU);

    if (ms->BoundsU.WLon > ms->CtrLonU)
	start -= DEG360_LU;

    end = ms->BoundsU.ELon;

    if (ms->BoundsU.ELon < ms->CtrLonU)
	end += DEG360_LU;

    for (xu = start; xu <= end; xu += GridU) {
	yu = 0;
	x1 = -999;
	for (yu = ms->BoundsU.SLat; yu < (ms->BoundsU.NLat + ThinU); yu += ThinU) {
	    if (yu > ms->BoundsU.NLat)
		yu = ms->BoundsU.NLat;
	    Project(ms, yu, xu, &x2, &y2);
	    if (x1 != -999) {
		ImageLine(image, x1, y, x2, y2, gc);
	    }
	    x1 = x2;
	    y = y2;
	}
    }
}				/* drawgraticule */


/* draw outside of globe when edge is showing */
void DrawGlobe(Pixmap image, MapStruct * ms, GC gc)
{
    int j = 0, x, y;
    long lat, lon;
    lon = ms->BoundsU.WLon;
    for (lat = -DEG90_LU; lat <= DEG90_LU; lat += 5000000) {
	Project(ms, lat, lon, &x, &y);
	ms->pointlist[j].x = x;
	ms->pointlist[j++].y = y;
	assert(j<=MAXPOINTS);
    }
    lon = ms->BoundsU.ELon;
    for (lat = 85000000; lat > -DEG90_LU; lat -= 5000000) {
	Project(ms, lat, lon, &x, &y);
	ms->pointlist[j].x = x;
	ms->pointlist[j++].y = y;
	assert(j<=MAXPOINTS);
    }
    ImageFilledPolygon(image, ms->pointlist, j, gc);
}
